unit hns_rea2;
{fits reading and mouse action on window realsky}

{By han_kleijn@hnsky.org. (c) 2001, 2002, 2003, for non-commercial use only }

{See also restrictions using getdss.dll in readme.doc file of project pluto}
{For updates have a look to www.hnsky.org}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, hns_udss;

type
  TRealsky = class(TForm)
    Image1: TImage;
    procedure Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Image1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);

  private
    { Private declarations }
  public
    { Public declarations }

  end;

var
  Realsky: TRealsky;
  date_obs,ut,pltlabel,plateid,telescop,instrum,origin,time_obs:string;
  nrbits    : integer;
  subsamp: integer;{not always available. For normal DSS =1}

const
   bandpass:real=0;{from fits file}
   equinox:real=0;{from fits file}
   exposure:real=0;{from fits file}
   ra0 : real=0;
   dec0: real=0; {plate center values}
   crpix1: real=0;{reference pixel}
   crpix2: real=0;
   cdelt1: real=0;{deg/pixel for x}
   cdelt2: real=0;

function  prepare_ra5(rax:real; sep:char):string; {radialen to text  format 24d 00.0}
function  prepare_dec5(decx:real;sep:char):string; {radialen to text  format 90d 00 }
procedure savefits8(filen,filen2:string;adjust:boolean);{save to 8bits fits file, by Han Kleijn}
procedure getfits_histogram(filename2:string);{load and calculate histogram, adapt brightness and background settings}
procedure load_fits(filename2:string);{normal without determination histogram}

implementation

uses hns_rea1;

{$R *.DFM}

const
      wide2  : integer=100;
      height2: integer=100;{do not use reserved word height}

var
  TheFile3  : tfilestream;
  Reader    : TReader;
  header    : array[0..$4000] of char;
const   bufwide=5000*8; 
var
  fitsbuffer : array[0..bufwide] of byte;{buffer for 8 bit FITS file}
  fitsbuffer2: array[0..round(bufwide/2)] of word absolute fitsbuffer;{buffer for 16 bit FITS file}
  fitsbuffer4: array[0..round(bufwide/4)] of longword absolute fitsbuffer;{buffer for floating bit ( -32) FITS file}
  fitsbuffer8: array[0..round(bufwide/8)] of int64 absolute fitsbuffer;{buffer for floating bit ( -64) FITS file}

  angle     : real; {image rotation at center, radians}
  cd1_1,cd1_2,cd2_1,cd2_2 :real;
  ram,decm  : real; {mouse position}
var
 histogram : array[0..255] of integer;
 datamin     : integer;
 datamax     : integer;



Function LeadingZero(w : integer) : String;
 var
   s : String;
 begin
   Str(w:0,s);
   if Length(s) = 1 then
     s := '0' + s;
   LeadingZero := s;
 end;
function prepare_ra5(rax:real; sep:char):string; {radialen to text  format 24d 00.0}
  var
    B : String[2];
    h,m,dm  :integer;
begin {make from rax [0..pi*2] a text in array bericht. Length is 8 long}
  rax:=rax+pi*0.1/(24*60); {add 1/10 of half minute to get correct rounding and not 7:60 results as with round}
  rax:=rax*12/pi; {make hours}
  h:=trunc(rax);
  m:=trunc((rax-h)*60);
  dm:=trunc((rax-h-m/60)*600);
  Str(trunc(h):2,b);
  prepare_ra5:=b+sep+' '+leadingzero(m)+'.'+char(dm+48);
end;

function prepare_dec5(decx:real;sep:char):string; {radialen to text  format 90d 00 }
 var
   B : String[7];
   g,m :integer;
   sign   : char;
begin {make from rax [0..pi*2] a text in array bericht. Length is 10 long}
  if decx<0 then sign:='-' else sign:='+';
  decx:=abs(decx)+pi/(360*60); {add half minute to get correct rounding and not 7:60 results as with round}
  decx:=decx*180/pi; {make degrees}
  g:=trunc(decx);
  m:=trunc((decx-g)*60);
  Str(trunc(g):0,b);
  prepare_dec5:=sign+b+sep+' '+leadingzero(m);
end;

function prepare_ra(rax:real; sep:char):string; {radialen to text, format 24: 00 00.0 }
 var
   B : String[2];
   h,m,s,ds  :integer;
 begin   {make from rax [0..pi*2] a text in array bericht. Length is 8 long}
  rax:=rax+pi*0.1/(24*60*60); {add 1/10 of half second to get correct rounding and not 7:60 results as with round}
  rax:=rax*12/pi; {make hours}
  h:=trunc(rax);
  m:=trunc((rax-h)*60);
  s:=trunc((rax-h-m/60)*3600);
  ds:=trunc((rax-h-m/60-s/3600)*36000);
  Str(trunc(h):2,b);
  prepare_ra:=b+sep+' '+leadingzero(m)+'  '+leadingzero(s)+'.'+char(ds+48);
end;

function prepare_dec(decx:real; sep:char):string; {radialen to text, format 90d 00 00}
 var
   B : String[9];
   g,m,s  :integer;
   sign   : char;
begin {make from rax [0..pi*2] a text in array bericht. Length is 10 long}
  if decx<0 then sign:='-' else sign:='+';
  decx:=abs(decx)+pi/(360*60*60); {add half second to get correct rounding and not 7:60 results as with round}
  decx:=decx*180/pi; {make degrees}
  g:=trunc(decx);
  m:=trunc((decx-g)*60);
  s:=trunc((decx-g-m/60)*3600);
  Str(trunc(g):2,b);
  prepare_dec:=sign+b+sep+' '+leadingzero(m)+'  '+leadingzero(s);
end;
function prepare_ra2(rax:real; sep:char):string; {radialen to text, format 24: 00 00.00 }
 var
   B       : String[2];
   h,m,s,ds  :integer;
 begin   {make from rax [0..pi*2] a text in array bericht. Length is 8 long}
  rax:=rax+pi*0.01/(24*60*60); {add 1/10 of half second to get correct rounding and not 7:60 results as with round}
  rax:=rax*12/pi; {make hours}
  h:=trunc(rax);
  m:=trunc((rax-h)*60);
  s:=trunc((rax-h-m/60)*3600);
  ds:=trunc((rax-h-m/60-s/3600)*360000);
  Str(trunc(h):2,b);
  prepare_ra2:=b+sep+' '+leadingzero(m)+'  '+leadingzero(s)+'.'+leadingzero(ds);
end;

Function prepare_dec2(decx:real; sep:char):string; {radialen to text, format 90d 00 00.1}
 var
   B,ds2 : String[5];
   g,m,s,ds :integer;
   sign   : char;
begin {make from rax [0..pi*2] a text in array bericht. Length is 10 long}
  if decx<0 then sign:='-' else sign:='+';
  decx:=abs(decx)+pi*0.1/(360*60*60); {add 1/20 second to get correct rounding and not 7:60 results as with round}
  decx:=decx*180/pi; {make degrees}
  g:=trunc(decx);
  m:=trunc((decx-g)*60);
  s:= trunc((decx-g-m/60)*3600);
  ds:=trunc((decx-g-m/60-s/3600)*36000);
  Str(trunc(g):2,b);
  Str(trunc(ds):1,ds2);
  prepare_dec2:=sign+b+sep+' '+leadingzero(m)+'  '+leadingzero(s)+'.'+ds2;
end;



procedure rotate(rot,x,y :real;var  x2,y2:real);{rotate a vector point}
begin
  x2:=x * + sin(rot)+y*cos(rot );{ROTATION MOON AROUND CENTER OF PLANET}
  y2:=x * - cos(rot)+y*sin(rot );{SEE PRISMA WIS VADEMECUM BLZ 68}
end;

FUNCTION fnmodulo (x,range: real):real;
BEGIN
   {range should be 2*pi or 24 hours or 0 .. 360}
   x:=range *frac(X /range); {quick method for big numbers}
   if x<0 then x:=x+range;   {do not like negative numbers}
   fnmodulo:=x;
END;

function intensity2(x:tcolor):integer;
begin
  intensity2:=round((GetBValue(x)+getGvalue(x)+getRvalue(x))/3);{get red, green blue value as intensity}
end;

//procedure equalize; Experimental on screen pixels. Not used
//var
// histogram : array[0..255] of integer;
// wrk       : array[0..255] of integer;
// wrk2      : array[0..255] of integer;
// row ,col ,i,j,p : integer;
//begin
//   for i := 0 to 255 do histogram[i] := 0;
//   for i:=0 to height2-1 do
//   begin
//    for j:=0 to wide2-1 do
//    begin
//      p:=intensity2(getpixel(realsky.image1.canvas.handle,j,i));
//      histogram[p]:=histogram[p]+1;
//    end;
//  end;
//
//  {equalize}
//  p:=0;
//  for i := 0 to 255 do
//  begin
//    p:=p+round(histogram[i]);
//    wrk[i] := round((255*p)/(height2*wide2));
//  end;

//// alternative for i := 1 to 255 do inc(histogram[i],histogram[i-1]);
//// alternative for i := 0 to 255 do wrk2[i] :=trunc(Histogram[i]/Histogram[255]*255);

//   for i:=0 to height2-1 do
//   begin
//     for j:=0 to wide2-1 do
//     begin
//       p:=wrk[intensity2(getpixel(realsky.image1.canvas.handle,j,i))];
//       setpixel(realsky.image1.canvas.handle,j,i,rgb(p,p,p));
//     end;
//  end;
//  realsky.Image1.Invalidate;
//end;





procedure plot_north;{draw arrow north}
var  x,y,x2,y2 :real;
const xpos=20;{position arrow}
      ypos=20;
      leng=14;{half of length}
begin
  realsky.image1.Canvas.Pen.Color := clWhite;

  rotate(angle,leng,0,x2,y2);
  moveToex(realsky.image1.Canvas.handle,round(xpos-x2),round(ypos-y2),nil);
  lineTo(realsky.image1.Canvas.handle,round(xpos+x2),round(ypos+y2));
  rotate(-30*pi/180+angle,8,0,x,y);
  lineTo(realsky.image1.Canvas.handle,round(xpos+x2-x),round(ypos+y2-y));
  rotate(+30*pi/180+angle,8,0,x,y);
  lineTo(realsky.image1.Canvas.handle,round(xpos+x2-x),round(ypos+y2-y));
  lineTo(realsky.image1.Canvas.handle,round(xpos+x2),round(ypos+y2));
end;

Function INT_IEEE4(x:longword):real;{adapt non-intel floating to intel floating point}
var lw       : longword;
    fl       : single absolute lw;
begin
  lw:=(x and $FF) shl 24 + (x and $FF00) shl 8 + (x and $FF0000) shr 8 +(x and $FF000000) shr 24 ;
      {swap 4 bytes such that 1234 becomes 4321}
  INT_IEEE4:=fl;{get final floating value}
end;
Function SWAP32fast(x:longword):longword;{it very little faster then SWAP32 and SWAP32first,  adapt non-intel 32  bit  longword to intel longword   or big endian to little endian}
asm
  bswap eax
end;

Function INT_IEEE8(x: int64):double;{adapt non-intel floating to intel floating point}
var lw       : int64;
    fl       : double absolute lw;
begin
  lw:=(x and $FF) shl 56 + (x and $FF00) shl 40 + (x and $FF0000) shl 24 + (x and $FF000000) shl 8 +(x and $FF00000000) shr 8 +(x and $FF0000000000) shr 24 + (x and $FF000000000000) shr 40 +(x and $FF00000000000000) shr 56;
      {swap 8 bytes such that 12345678 becomes 87654321}
  INT_IEEE8:=fl;{get final floating value}
end;
procedure adjustdataminmax(datamin,datamax:integer;var datamin2,datamax2:integer);{adjust datamin/max to brightness and background settings}
begin
  datamin2:=datamin+round((datamax-datamin)*mainwindow.background.position/100);
  datamax2:=round(datamax*(1-mainwindow.brightness.position/100));
  if datamax2<=datamin2 then datamax2:=datamin2+1;
end;

Procedure loadfits(filen:string;histogr:boolean);{load fits file, by Han Kleijn}
var
   i,j,sign    : integer;
   col         : integer;
   crota2,crota1, fract: real; {image rotation at center in degrees}
   nr,error3   : integer;
   s           : string[3];
   bzero       : integer;{zero shift. For example used in AMT, Tricky do not use int64,  maxim DL writes BZERO value -2147483647 as +2147483648 !! }

   wo          : word;   {for 16 signed integer}
   sign_int    : smallint absolute wo;{for 16 signed integer}


type
  PRGBTripleArray = ^TRGBTripleArray; {for fast pixel routine}
  TRGBTripleArray = array[0..4095] of TRGBTriple; {for fast pixel routine}
var
   Bitmap  : TBitmap;{for fast pixel routine}
   pixelrow: PRGBTripleArray;{for fast pixel routine}


     Function validate_real:real;{read values}
     var t :shortstring;
         r,err : integer;
         x     :real;
     begin
       t:='';
       for r:=I+10 to I+29+1 do {read 20 characters max, position 11 to 30 in string, position 10 to 29 in pchar}
       if header[r] in [' ','/']=false then t:=t+header[r];{'/' is strictly not neccesary but safer}
       val(t,x,err);
       validate_real:=x;
     end;
     Function get_string:string;{read values}
     var t :shortstring;
         r : integer;
     begin
       t:='';
       r:=I+10+1;
       repeat
         t:=t+header[r];
         inc(r);
       until ((header[r]=#39) or (r>=I+10+69));
       get_string:=t;
     end;

     Procedure old_to_new_WCS;{ convert old WCS to new}
     begin
       cd1_1:=+cdelt1 * cos(crota1*pi/180);  {2013 included crota1 instead of crota2}
       cd1_2:=-cdelt2 * sin(crota1*pi/180);
       cd2_1:=+cdelt1 * sin(crota2*pi/180);
       cd2_2:=+cdelt2 * cos(crota2*pi/180);
     end;

     Function get_angle:real;{ get angle from PC matrix}
     var
       sign        : integer;
     begin
       if (cd1_1*cd2_2-cd1_2*cd2_1)>=0 then sign:=+1 else sign:=-1;
       get_angle:= -atn_2(sign*cd1_2,cd2_2);{for North arrow}
     end;
begin
  realsky.caption:=ExtractFileName(filen);
  try
    TheFile3:=tfilestream.Create( filen, fmOpenRead );
  except
     beep;
     textout(realsky.image1.canvas.handle,30,30,'Error loading FITS file !',25);
     exit;
  end;


  Reader := TReader.Create (theFile3,$4000);{number of hnsky records}
  {thefile3.size-reader.position>sizeof(hnskyhdr) could also be used but slow down a factor of 2 !!!}
  I:=0;
  crota2:=0;{just for the case it is not available}
  crota1:=0;
  ra0:=0;
  dec0:=0;
  subsamp:=1;{just for the case it is not available}
  bzero:=0;{just for the case it is not available}
  cd1_1:=0;{just for the case it is not available}
  cd1_2:=0;{just for the case it is not available}
  cd2_1:=0;{just for the case it is not available}
  cd2_2:=0;{just for the case it is not available}
  date_obs:=''; ut:=''; pltlabel:=''; plateid:=''; telescop:=''; instrum:='';  origin:=''; time_obs:='';{clear}

  if histogr=true then for i := 0 to 255 do histogram[i] := 0;{clear histogram}

  repeat
    try reader.read(header[I],5*16);except; end; {read file info, 5*16 bytes only}
    if ((header[i]='B') and (header[i+1]='I')  and (header[i+2]='T') and (header[i+3]='P') and (header[i+4]='I') and (header[i+5]='X')) then
    begin
      nrbits:=round(validate_real);{BITPIX, read integer using real routine}
      if nrbits=8 then  begin datamin:=0;datamax:=255; {8 bits files} end
      else {16, -32 files}
      begin datamin:=0;datamax:=$FFFF;end{not always specified. For example in skyview. So refresh here for case brightness is adjusted}


    end;
    if ((header[i]='N') and (header[i+1]='A')  and (header[i+2]='X') and (header[i+3]='I') and (header[i+4]='S')) then
       begin
         if (header[i+5]='1') then wide2:=round(validate_real) else {NAXIS1 pixels}
         if (header[i+5]='2') then height2:=round(validate_real);   {NAXIS2 pixels}
       end;
    if ((header[i]='D') and (header[i+1]='A')  and (header[i+2]='T') and (header[i+3]='A') and (header[i+4]='M')) then
       begin
         if mainwindow.use_datamin.checked then
         begin
           if ((header[i+5]='I') and (header[i+6]='N')) then datamin:=round(validate_real) else {DATAMIN. Note some FIT files contain real}
           if ((header[i+5]='A') and (header[i+6]='X')) then datamax:=round(validate_real);{DATAmax. Note some FIT files contain real }
         end;
       end;
    if ((header[i]='B') and (header[i+1]='Z')  and (header[i+2]='E') and (header[i+3]='R') and (header[i+4]='O') ) then
             bzero:=round(validate_real);

    if ((header[i]='C') and (header[i+1]='R')  and (header[i+2]='O') and (header[i+3]='T') and (header[i+4]='A') and (header[i+5]='2')) then
             crota2:=validate_real;{DATAMIN, read real value}                                                    {last for r Ledas}

    if ((header[i]='C') and (header[i+1]='R')  and (header[i+2]='O') and (header[i+3]='T') and (header[i+4]='A') and (header[i+5]='1')) then
             crota1:=validate_real;{DATAMIN, read real value}

    if ((header[i]='C') and (header[i+1]='R')  and (header[i+2]='P') and (header[i+3]='I') and (header[i+4]='X')) then
        begin
          if header[i+5]='1' then crpix1:=validate_real else{ref pixel for x}
          if header[i+5]='2' then crpix2:=validate_real;    {ref pixel for y}
        end;
    if ((header[i]='C') and (header[i+1]='D')  and (header[i+2]='E') and (header[i+3]='L') and (header[i+4]='T')) then
        begin
          if header[i+5]='1' then cdelt1:=validate_real else{deg/pixel for RA}
          if header[i+5]='2' then cdelt2:=validate_real;    {deg/pixel for DEC}
        end;

    if ((header[i]='C') and (header[i+1]='R')  and (header[i+2]='V') and (header[i+3]='A') and (header[i+4]='L')) then
        begin
          if (header[i+5]='1') then  ra0:=validate_real*pi/180; {ra center, read real value}
          if (header[i+5]='2') then  dec0:=validate_real*pi/180; {dec center, read real value}
        end;
    if ((header[i]='C') and (header[i+1]='D')) then
        begin
          if ((header[i+2]='1') and (header[i+3]='_') and (header[i+4]='1')) then   cd1_1:=validate_real;
          if ((header[i+2]='1') and (header[i+3]='_') and (header[i+4]='2')) then   cd1_2:=validate_real;
          if ((header[i+2]='2') and (header[i+3]='_') and (header[i+4]='1')) then   cd2_1:=validate_real;
          if ((header[i+2]='2') and (header[i+3]='_') and (header[i+4]='2')) then   cd2_2:=validate_real;
        end;


    {following is only required when using DSS polynome plate fit}
    if ((header[i]='A') and (header[i+1]='M')  and (header[i+2]='D') and (header[i+3]='X')) then
        begin
          if header[i+5]=' ' then s:=(header[i+4]) else s:=(header[i+4])+(header[i+5]);
          val(s,nr,error3);{1 to 20}
          x_coeff[nr-1]:=validate_real;
        end;
    if ((header[i]='A') and (header[i+1]='M')  and (header[i+2]='D') and (header[i+3]='Y')) then
        begin
          if header[i+5]=' ' then s:=(header[i+4]) else s:=(header[i+4])+(header[i+5]);
          val(s,nr,error3);{1 to 20}
          y_coeff[nr-1]:=validate_real;
        end;
    if ((header[i]='P') and (header[i+1]='P')  and (header[i+2]='O')) then
        begin
          if (header[i+3]='3') then   ppo_coeff[2]:=validate_real;
          if (header[i+3]='6') then   ppo_coeff[5]:=validate_real;
        end;
    if ((header[i]='C') and (header[i+1]='N')  and (header[i+2]='P') and (header[i+3]='I') and (header[i+4]='X') and (header[i+5]='1')) then
              x_pixel_offset:=round(validate_real);{rotation, read real value}
    if ((header[i]='C') and (header[i+1]='N')  and (header[i+2]='P') and (header[i+3]='I') and (header[i+4]='X') and (header[i+5]='2')) then
              y_pixel_offset:=round(validate_real);{rotation, read real value}
    if ((header[i]='X') and (header[i+1]='P')  and (header[i+2]='I') and (header[i+3]='X') and (header[i+4]='E') and (header[i+5]='L')) then
              x_pixel_size:=validate_real;{rotation, read real value}
    if ((header[i]='Y') and (header[i+1]='P')  and (header[i+2]='I') and (header[i+3]='X') and (header[i+4]='E') and (header[i+5]='L')) then
              y_pixel_size:=validate_real;{rotation, read real value}
    if ((header[i]='P') and (header[i+1]='L')  and (header[i+2]='T') and (header[i+3]='R') and (header[i+4]='A')) then
        begin
          if (header[i+5]='H') then   plate_ra:=validate_real*pi/12;
          if (header[i+5]='M') then   plate_ra:=plate_ra+validate_real*pi/(60*12);
          if (header[i+5]='S') then   plate_ra:=plate_ra+validate_real*pi/(60*60*12);;
        end;
    if ((header[i]='P') and (header[i+1]='L')  and (header[i+2]='T') and (header[i+3]='D') and (header[i+4]='E')) then
        begin
          if (header[i+7]='N') then  begin if (header[i+11]='-') then  dec_sign:=-1 else dec_sign:=+1;end; {dec sign}
          if (header[i+6]='D') then   plate_dec:=validate_real*pi/180;
          if (header[i+6]='M') then   plate_dec:=plate_dec+validate_real*pi/(60*180);
          if (header[i+6]='S') then   plate_dec:=dec_sign*(plate_dec+validate_real*pi/(60*60*180));
        end;
    if ((header[i]='S') and (header[i+1]='U')  and (header[i+2]='B') and (header[i+3]='S') and (header[i+4]='A') and (header[i+5]='M')) then
              subsamp:=round(validate_real);{subsampling value}
    {end using DSS polynome plate fit}



    if ((header[i]='E') and (header[i+1]='Q')  and (header[i+2]='U') and (header[i+3]='I') and (header[i+4]='N') and (header[i+5]='O') and (header[i+6]='X')) then
             equinox:=validate_real;{read real value}
    if ((header[i]='E') and (header[i+1]='X')  and (header[i+2]='P') and (header[i+3]='O') and (header[i+4]='S') and (header[i+5]='U') and (header[i+6]='R')) then
             exposure:=validate_real;{read real value}
    if ((header[i]='B') and (header[i+1]='A')  and (header[i+2]='N') and (header[i+3]='D') and (header[i+4]='P') and (header[i+5]='A') and (header[i+6]='S')) then
             BANDPASS:=validate_real;{read integer as real value}

    if ((header[i]='D') and (header[i+1]='A')  and (header[i+2]='T') and (header[i+3]='E') and (header[i+4]='-') and (header[i+5]='O') and (header[i+6]='B')) then
             DATE_OBS:=get_string;
    if ((header[i]='U') and (header[i+1]='T')) then
             UT:=get_string;
    if ((header[i]='P') and (header[i+1]='L')  and (header[i+2]='T') and (header[i+3]='L') and (header[i+4]='A') and (header[i+5]='B') and (header[i+6]='E')) then
             PLTLABEL:=get_string;
    if ((header[i]='P') and (header[i+1]='L')  and (header[i+2]='A') and (header[i+3]='T') and (header[i+4]='E') and (header[i+5]='I') and (header[i+6]='D')) then
             PLATEID:=get_string;
    if ((header[i]='T') and (header[i+1]='E')  and (header[i+2]='L') and (header[i+3]='E') and (header[i+4]='S') and (header[i+5]='C') and (header[i+6]='O')) then
             TELESCOP:=get_string;
    if ((header[i]='O') and (header[i+1]='R')  and (header[i+2]='I') and (header[i+3]='G') and (header[i+4]='I') and (header[i+5]='N')) then
             origin:=get_string;
    if ((header[i]='I') and (header[i+1]='N')  and (header[i+2]='S') and (header[i+3]='T') and (header[i+4]='R') and (header[i+5]='U') and (header[i+6]='M')) then
             INSTRUM:=get_string;
    if ((header[i]='T') and (header[i+1]='I')  and (header[i+2]='M') and (header[i+3]='E') and (header[i+4]='-') and (header[i+5]='O') and (header[i+6]='B')) then
             time_obs:=get_string;



   until (((header[i]='E') and (header[i+1]='N')  and (header[i+2]='D')) or (I>=sizeof(header)-16 ));

   header[i]:=#0; {create end in header }

   if (crota1=0) then crota1:=crota2; {for case crota1 is not used}

   if cdelt2=0 then {new WCS, calculate rotation for North arrow}
     angle:=get_angle {calculate angle form CD matrix}
   else  {old WCS}
   begin
     old_to_new_WCS;{ convert old WCS to new}
     angle:=(-crota2)*pi/180;{for North arrow}
   end;
   fract:=frac(reader.position/2880);
   if fract<>0 then
   begin
     i:=round((1-fract)*2880);{left part of next 2880 bytes block. For realsky read total 4 x 2880 blocks}
     try reader.read(header[0],i);except;end; {skip empty part and go to image data}
   end;
   i:=round(bufwide/(abs(nrbits/8)));
   if wide2>i then
   begin
     beep;
     textout(realsky.image1.canvas.handle,30,30,'Too large FITS file !!!!!',25);
     nrbits:=8; {do not allow saving to 8 bits }
   end
   else
   begin {not too large}
     realsky.clientwidth:=wide2; {minimum width}
     realsky.ClientHeight:=height2; {window size}

     realsky.image1.width:=wide2; {image size 2013}
     realsky.image1.height:=height2; {image size}

     if histogr=false then adjustdataminmax(datamin,datamax,datamin,datamax);{adjust datamin/max to brightness and background settings}



     Bitmap := TBitmap.Create;
     TRY
       WITH Bitmap DO
       BEGIN
        Width := wide2;
        Height := height2;
        // Unclear why this must follow width/height to work correctly.
        // If PixelFormat precedes width/height, bitmap will always be black.
        Bitmap.PixelFormat := pf24bit;
        END;

     For i:=0 to height2-1 do
     begin
       try reader.read(fitsbuffer,wide2*round(abs(nrbits/8)));except; end; {read file info}

       pixelrow := Bitmap.ScanLine[(height2-1)-i];{height2-1)-i, FITS count from bottom, windows from top}

       for j:=0 to wide2-1 do
       begin
         if nrbits=8  then col:=round( ((fitsbuffer[j]-datamin+bzero)*256/(datamax-datamin) )){scale to maximum contrast}
         else
         if nrbits=16 then
         begin
           wo:=swap(fitsbuffer2[j]);{move data to wo and therefore sign_int}
           col:=round( ((sign_int-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
         end
         else {4 byte floating point FITS image}
         if nrbits=-32 then col:=round( (( INT_IEEE4(fitsbuffer4[j])-datamin+bzero)*256/(datamax-datamin) )){scale to maximum contrast}
                        {int_IEEE, swap four bytes and the read as floating point}

         else {4 byte, 32 bit FITS image}
         if nrbits=+32 then
         begin
           col:=round( (( SWAP32fast(fitsbuffer4[j])-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
                        {int_IEEE, swap four bytes and the read as longword}
                        {Tricky do not use int64 for BZERO,  maxim DL writes BZERO value -2147483647 as +2147483648 !!}
         end

         else {8 byte floating point FITS image}
         if nrbits=-64 then col:=round( (( INT_IEEE8(fitsbuffer8[j])-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
                        {int_IEEE, swap four bytes and the read as floating point}



         if col>255 then col:=255;
         if col<0 then col:=0;
         if histogr=false then
                 begin
                  pixelrow[j].rgbtBlue := col;{fast pixel write routine }
                  pixelrow[j].rgbtGreen := col;
                  pixelrow[j].rgbtRed := col;
                 end
//         setpixel(realsky.image1.canvas.handle,j,(height2-1)-i,rgb(col,col,col)){too slow}
         else
         inc(histogram[col],1);{calculate histogram}
       end;
     end;
     if histogr=false then realsky.image1.picture.Graphic := Bitmap; {show image}

     FINALLY
       Bitmap.Free
     end;
   end; {not too large}

   Reader.free;
   TheFile3.free;

   mainwindow.Saveas8bitFITS1.enabled:=nrbits<>8; {allow saving to 8 bits only if}
   mainwindow.Saveas8bitFITSfileadjusted.enabled:=nrbits<>8; {allow saving to 8 bits only if}
   if mainwindow.north_arrow.checked then plot_north; {draw arrow indication position north}
   invalidaterect(realsky.handle,nil,false);{paint again}
 end;
procedure getfits_histogram(filename2:string);{load and calculate histogram, adapt brightness and background settings}
var
 i, min, max : integer;
begin
  loadfits(filename2,true); {load for histogram only}

  min:=0;
  max:=256;
  for i := 0 to 255 do
    if ((min=0  ) and (histogram[i]>0.01*(height2*wide2))) then min:=i; {find where image black values starts}
  for i := 255 downto 0 do
    if ((max=256) and (histogram[i] >0.0002 *(height2*wide2))) then max:=i; {find where image maximum values starts}

  mainwindow.background.position:=round(100*min/256);
  mainwindow.brightness.position:=round(100*(1-max/256));
end;
procedure load_fits(filename2:string);{normal without histogram calculation}
begin
  loadfits(filename2,false); {load now normal}
end;


procedure savefits8(filen,filen2:string;adjust:boolean);{save to 8bits fits file}
var
  datamin     : integer;
  datamax     : integer;
  TheFile4  : tfilestream;
  I,j,col   : integer;
  fract     : real;
  bzero       : integer;{zero shift. For example used in AMT}
  wo          : word;   {for signed 16 bit integer}
  sign_int    : smallint absolute wo;{for signed 16 bit integer}


     Function validate_real:real;{read values}
     var t :shortstring;
         r,err : integer;
         x     :real;
     begin
       t:='';
       for r:=I+10 to I+29 do
       if header[r]<>' ' then t:=t+header[r];
       val(t,x,err);
       validate_real:=x;
     end;
begin

  try
    TheFile3:=tfilestream.Create(filen, fmOpenRead );
  except
    TheFile3.free;
    exit;
  end;
  try
    TheFile4:=tfilestream.Create(filen2, fmcreate );
  except
    TheFile4.free;
    exit;
  end;

  Reader := TReader.Create (theFile3,$4000);{number of hnsky records}
  {thefile3.size-reader.position>sizeof(hnskyhdr) could also be used but slow down a factor of 2 !!!}
  I:=0;
  bzero:=0;{just for the case it is not available}
  subsamp:=0; {buf fix temporary}
  repeat
    try reader.read(header[I],5*16);except;end; {read file info, 5*16 bytes only}
    if ((header[i]='B') and (header[i+1]='I')  and (header[i+2]='T') and (header[i+3]='P') and (header[i+4]='I') and (header[i+5]='X')) then
    begin
      nrbits:=round(validate_real);
      if nrbits=8 then  begin datamin:=0;datamax:=255; {8 bits files} end
      else {16 , -32  files}
      begin datamin:=0;datamax:=$FFFF;end;{not always specified. For example in skyview. So refresh here for case brightness is adjusted}

      header[i+27]:=' ';  {change header to 8 bits}
      header[i+28]:=' ';
      header[i+29]:='8';
    end;
    if ((header[i]='N') and (header[i+1]='A')  and (header[i+2]='X') and (header[i+3]='I') and (header[i+4]='S')) then
       begin
         if (header[i+5]='1') then wide2:=round(validate_real) else {NAXIS1 pixels}
         if (header[i+5]='2') then height2:=round(validate_real);   {NAXIS2 pixels}
       end;
    if ((header[i]='D') and (header[i+1]='A')  and (header[i+2]='T') and (header[i+3]='A') and (header[i+4]='M') and (header[i+5]='I')) then
    begin
      if mainwindow.use_datamin.checked then datamin:=round(validate_real);{use or ignore datamin/datamax}
      header[i+10]:=' ';
      header[i+11]:=' ';
      header[i+12]:=' ';
      header[i+13]:=' ';
      header[i+14]:=' ';
      header[i+15]:=' ';
      header[i+16]:=' ';
      header[i+17]:=' ';
      header[i+18]:=' ';
      header[i+19]:=' ';
      header[i+20]:=' ';
      header[i+21]:=' ';
      header[i+22]:=' ';
      header[i+23]:=' ';
      header[i+24]:=' ';
      header[i+25]:=' ';
      header[i+26]:=' ';
      header[i+27]:=' ';
      header[i+28]:=' ';
      header[i+29]:='0';{datamin value}
    end;
    if ((header[i]='D') and (header[i+1]='A')  and (header[i+2]='T') and (header[i+3]='A') and (header[i+4]='M') and (header[i+5]='A')) then
    begin
      if mainwindow.use_datamin.checked then datamax:=round(validate_real);{use or ignore datamin/datamax}
      header[i+10]:=' ';
      header[i+11]:=' ';
      header[i+12]:=' ';
      header[i+13]:=' ';
      header[i+14]:=' ';
      header[i+15]:=' ';
      header[i+16]:=' ';
      header[i+17]:=' ';
      header[i+18]:=' ';
      header[i+19]:=' ';
      header[i+20]:=' ';
      header[i+21]:=' ';
      header[i+22]:=' ';
      header[i+23]:=' ';
      header[i+24]:=' ';
      header[i+25]:=' ';
      header[i+26]:=' ';
      header[i+27]:='2';
      header[i+28]:='5';
      header[i+29]:='5';{datamax value}
    end;
    if ((header[i]='B') and (header[i+1]='Z')  and (header[i+2]='E') and (header[i+3]='R') and (header[i+4]='O') ) then
    begin
      bzero:=round(validate_real);
      header[i+10]:=' ';
      header[i+11]:=' ';
      header[i+12]:=' ';
      header[i+13]:=' ';
      header[i+14]:=' ';
      header[i+15]:=' ';
      header[i+16]:=' ';
      header[i+17]:=' ';
      header[i+18]:=' ';
      header[i+19]:=' ';
      header[i+20]:=' ';
      header[i+21]:=' ';
      header[i+22]:=' ';
      header[i+23]:=' ';
      header[i+24]:=' ';
      header[i+25]:=' ';
      header[i+26]:=' ';
      header[i+27]:=' ';
      header[i+28]:=' ';
      header[i+29]:='0';{datamax value}
    end;
    if ((header[i]='S') and (header[i+1]='W')  and (header[i+2]='M') and (header[i+3]='O') and (header[i+4]='D') ) then
    begin
      header[i+10]:=' '; {remove software modifications labels}
      header[i+11]:=' ';
      header[i+12]:=' ';
      header[i+13]:=' ';
      header[i+14]:=' ';
      header[i+15]:=' ';
      header[i+16]:=' ';
      header[i+17]:=' ';
      header[i+18]:=' ';
      header[i+19]:=' ';
      header[i+20]:=' ';
      header[i+21]:=' ';
      header[i+22]:=' ';
      header[i+23]:=' ';
      header[i+24]:=' ';
      header[i+25]:=' ';
      header[i+26]:=' ';
      header[i+27]:=' ';
      header[i+28]:=' ';
      header[i+29]:=' '; header[i+30]:=' ';  header[i+31]:=' ';   header[i+32]:=' ';    header[i+33]:=' ';
      header[i+34]:=' '; header[i+35]:=' ';  header[i+36]:=' ';   header[i+37]:=' ';    header[i+38]:=' ';
      header[i+39]:=' '; header[i+40]:=' ';  header[i+41]:=' ';   header[i+42]:=' ';    header[i+43]:=' ';
      header[i+44]:=' '; header[i+45]:=' ';  header[i+46]:=' ';   header[i+47]:=' ';    header[i+48]:=' ';
      header[i+49]:=' ';
    end;

   thefile4.write(header[I],5*16);
   until (((header[i]='E') and (header[i+1]='N')  and (header[i+2]='D')) or (I>=sizeof(header)-16 ));

   header[i]:=#0; {create end in header }

   fract:=frac(reader.position/2880);
   if fract<>0 then
   begin
     i:=round((1-fract)*2880);{left part of next 2880 bytes block. For realsky read total 4 x 2880 blocks}
     try reader.read(header[0],i);except;end; {skip empty part and go to image data}
     thefile4.write(header[0] {fitsbuffer2},i);
   end;
   if adjust then adjustdataminmax(datamin,datamax,datamin,datamax);{adjust datamin/max to brightness and background settings}

   i:=round(bufwide/(abs(nrbits/8)));
   if wide2>i then
   begin
      nrbits:=0; {too large image, conversion not possible delete later}
   end
   else
   for i:=0 to height2-1 do
   begin
     try reader.read(fitsbuffer,wide2*round(abs(nrbits/8)));except;end; {read file info}
     for j:=0 to wide2-1 do
     begin
       if nrbits=16 then
       begin
         wo:=swap(fitsbuffer2[j]);{move data to wo and therefore sign_int}
         col:=round( ((sign_int-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
       end
       else {4 byte floating point FITS image}
       if nrbits=-32 then col:=round( (( INT_IEEE4(fitsbuffer4[j])-datamin+bzero)*256/(datamax-datamin) )){scale to maximum contrast}
                        {int_IEEE, swap four bytes and the read as floating point}
       else
       if nrbits=-64 then col:=round( (( INT_IEEE8(fitsbuffer8[j])-datamin+bzero)*256/(datamax-datamin) ));{scale to maximum contrast}
                        {int_IEEE, swap four bytes and the read as floating point}


       if col>255 then col:=255;
       if col<0 then col:=0;
       fitsbuffer[j]:=col;
     end;
     thefile4.write(fitsbuffer,wide2); {write as bytes}
   end;

   Reader.free;
   TheFile3.free;
   TheFile4.free;
   if ((nrbits=8) or (nrbits=0)) then  {errors}
   begin
     DeleteFile(filen2); {do not convert 8 to 8 bit}
     if nrbits=8 then
     messagebox(mainwindow.handle, 'Already an 8 bit FITS file !', pchar(filen),MB_OK+MB_ICONWARNING)
     else
     messagebox(mainwindow.handle, 'Too large FITS file !', pchar(filen),MB_OK+MB_ICONWARNING);
     nrbits:=8;
   end;
end;


procedure TRealsky.Image1MouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  xx1,yy1,yy,xx,arcsinx, dRa,dDec,delta,gamma : real;
  error2 :real;
  s1,s2,s3  : string;
  fitsx,fitsy,fits_unsampledX, fits_unsampledY :real;
begin
  fitsx:=0.5+(0.5+x)/(image1.width/wide2);{starts at 1}
  fitsy:=0.5+height2-(0.5+y)/(image1.height/height2); {from bottom to top, starts at 1}

  fitsx:=round((image1.width/wide2)*fitsx)/(image1.width/wide2);
  fitsy:=round((image1.height/height2)*fitsy)/(image1.height/height2);

  if mainwindow.north_arrow.checked then
  begin
    str(fitsx:3:3,s1);  {fits images start with 1 and not with 0}
    str(fitsy:3:3,s2); {Y from bottom to top}
    try str( 1000+ (getpixel(realsky.canvas.handle,x,y) and $FF)  :4 ,s3);except;end;{1000 to get zeros, remove 1 later}
    {note  getpixel(image1.canvas.handle,x,y) doesn't work well since X,Y follows zoom  factor !!!}

    s1:=copy(s3,2,3)+'=('+s1+','+s2+')    '{+  floattostr(x)+'  '+floattostr(y)};
  end
  else s1:=' ';

  error2:=0;
 {DSS polynom solution}
  if mainwindow.dsspolynom.checked=true then
  begin
   { Convert from image subsampled pixels position to unsampled pixel position }
    fits_unsampledX:=subsamp*(fitsX-0.5)+0.5;
    fits_unsampledY:=subsamp*(fitsY-0.5)+0.5;
                  //{fits (1,1)+subsamp of 2x =>(eqv unsampled 1,5,1,5)
                  // fits (2,2)+subsamp of 2x =>(eqv unsampled 3,5,3,5)
                  //(fits 1,1)+subsamp of 4x=>(eqv unsampled 2.5,2.5)
                  //(fits 2,2)+subsamp of 4=>(eqv unsampled 6.5,6.5)
    dsspos (fits_unsampledX , fits_unsampledY, ram, decm );
  end
  else
  begin {WCS solution}
//  if cd1_1=0 then {simplified WCS no cd1_1... available}
//  begin
//    xx1:=  (pi/(60*180))*(60*-cdelt1)*(fitsx- crpix1 );
//    yy1:=  (pi/(60*180))*(60*+cdelt2)*(fitsy- crpix2 );{rotation factors}

      {correct rotation of plate, see 12.2 "astronomy on the personal computer" }
//    xx:=xx1*cos(angle)-yy1*sin(angle);
//    yy:=xx1*sin(angle)+yy1*cos(angle);{can be signifcant at poles !!!}

     {plate XX,YY coordinates to RA, DEC}
     {RA0,DEC0: right ascension and declination of the optical axis (deg) }
     {XX,YY:  plate coordinates in north/west}
     {RAM,DECM:   right ascension and declination}
     {see "astronomy on the personal computer" formula 12.2}
     {alternative DECM := arcsiN ( (sin(dec0)+y*cos(dec0))/SQRT(1.0+X*X+Y*Y) );{ArcSin(x) = ArcTan (x/sqrt (1-sqr (x)))}
//    RAM  := ra0 + atn_2( -XX , (cos(DEC0)- YY*sin(DEC0)) );
//    arcsinx:=(sin(dec0)+YY*cos(dec0))/SQRT(1.0+XX*XX+YY*YY);
//    DECM:=atn_2( arcsinx ,sqrt (1-sqr (arcsinx)) );
//    RAM:=fnmodulo (RAM,2*pi);
//  end
//  else
    begin  {improved new WCS}
      dRa :=(cd1_1*(fitsx-crpix1)+cd1_2*(fitsy-crpix2))*pi/180;
      dDec:=(cd2_1*(fitsx-crpix1)+cd2_2*(fitsy-crpix2))*pi/180;
      delta:=cos(dec0)-dDec*sin(dec0);
      gamma:=sqrt(dRa*dRa+delta*delta);
      RAM:=RA0+arctan(Dra/delta);
      DECM:=arctan((sin(dec0)+dDec*cos(dec0))/gamma);
    end;

  end;{WCS solution}
  if error2=0 then
  begin
//    statusbar1.simpletext:=s1+prepare_ra2(raM,':')+'   '+prepare_dec2(decM,'d');
    realsky.caption:=ExtractFileName(filename2)+'  '+s1+prepare_ra2(raM,':')+'   '+prepare_dec2(decM,'d');
  end
  else realsky.caption:=ExtractFileName(filename2)+'    Error DSS conversion';
end;

procedure TRealsky.Image1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var TheData : array[0..23] of ansichar;
      s1,s2 : string[12];
begin

  if ssright in shift then
  begin
    {Center on mouse click}
    realsky.caption:='Centering on this mouse position';
    mainwindow.edit1.text:=prepare_ra5(raM,':');
    mainwindow.edit2.text:=prepare_dec5(decM,':');
    mainwindow.Button1Click(sender)
  end;

  if ssleft in shift then {send position to hnsky}
  begin
     realsky.caption:='Sending this position by DDE';
    {The length of the DDE string send by HNSKY is "1.6,+1.6" total 18, where the}
    {RA has a length of 8 and DEC 9 charactors.}
    {RA and DEC are separted by a komma. }
    {Example: 3.141593,-1.123456 or 3.141593, 1.123456}
    str(raM  :8:6,s1);
    str(DECM :9:6,s2);

    {HNS_REAL client personality}
    strpcopy(TheData,s1+','+s2);
    mainwindow.DDEClientConv1.PokeData(mainwindow.DDEClientItem2.DDEItem, TheData);
    {poke data to HNSKY server item telescope}

    {HNS_REAL server personality}
    str(raM*12/pi:0:6,s1);{here direct after plot grid to get latest ra,dec}
    str(decM*180/pi:9:6,s2);
    mainwindow.mouse_pos.text:=s1+','+s2; {save dde_text for later use. Contains is lost after paint !!!}{dynamic date exchange or DDE items}
  end;
end;

procedure TRealsky.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  mainwindow.ShowFITSheader1.enabled:=false;
  mainwindow.SaveasBMP1.enabled:=false;
end;

procedure TRealsky.FormShow(Sender: TObject);
begin
  mainwindow.ShowFITSheader1.enabled:=true;
  mainwindow.SaveasBMP1.enabled:=true;
end;

{$R hns_realcursor.res}   {FOR CURSORS}
const
  crMyCursor = 5;
procedure TRealsky.FormCreate(Sender: TObject);
begin
  Screen.Cursors[crMyCursor] := LoadCursor(HInstance, 'cross_cursor');
  Cursor := crMyCursor;
end;

procedure TRealsky.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
   if realsky.ClientHeight>=100 then
   begin
     if (realsky.image1.height<=realsky.ClientHeight*2 ) then {same size, keep on resizing}
     begin
        realsky.clientwidth:=round(realsky.image1.width*0.5);
        realsky.ClientHeight:=round(realsky.image1.Height*0.5) ; {window size}
     end;
     realsky.image1.height:=round(realsky.image1.height*0.5);
     realsky.image1.width:=round(realsky.image1.width*0.5);

   end;

 end;

procedure TRealsky.FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
   if realsky.image1.height<=8000 then
   begin
     if ({(realsky.clientheight=realsky.image1.height) or}  (realsky.clientheight<450) ) then {same size, keep on resizing}
     begin
       realsky.clientwidth:=round(realsky.image1.width*2);
       realsky.ClientHeight:=round(realsky.image1.height*2); {window size}
     end;
     realsky.image1.height:=round(realsky.image1.height*2);
     realsky.image1.width:=round(realsky.image1.width*2);

   end;
end;


end.
